# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from featureprobe.internal.json_decoder import json_decoder
from featureprobe.model.segment import Segment
from featureprobe.model.toggle import Toggle


class Repository:
    def __init__(self,
                 toggles: Dict[str, "Toggle"] = None,
                 segments: Dict[str, "Segment"] = None):
        self._toggles = toggles or {}
        self._segments = segments or {}

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Repository":
        toggles = json.get('toggles', {})
        segments = json.get('segments', {})
        return cls(
            toggles={k: Toggle.from_json(v) for k, v in toggles.items()},
            segments={k: Segment.from_json(v) for k, v in segments.items()}
        )

    @property
    def toggles(self) -> Dict[str, "Toggle"]:
        return self._toggles

    @toggles.setter
    def toggles(self, value: Dict[str, "Toggle"]):
        self._toggles = value or {}

    @property
    def segments(self) -> Dict[str, "Segment"]:
        return self._segments

    @segments.setter
    def segments(self, value: Dict[str, "Segment"]):
        self._segments = value or {}



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def stringifiable(cls):
    """Experimental, unusable yet"""

    def __str__(self):
        attrs = sorted(filter(lambda attr: not callable(attr[1]),
                              self.__dict__.items()))

        return '%s(%s)' % (
            cls.__name__,
            ', '.join(
                '%s=\'%s\'' % item
                if isinstance(item[1], (str, bytes))
                else '%s=%s' % item
                for item in attrs
            )
        )

    cls.__str__ = __str__
    return cls



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Union


def peer_cmp_only(func):
    def type_check(self, other):
        return func(self, other) if isinstance(other, SemVer) else False

    return type_check


class SemVer:
    """Under the [Semantic Versioning 2.0.0](https://semver.org/spec/v2.0.0.html) standard."""

    __slots__ = ('major', 'minor', 'patch', 'pre_release')

    # https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
    FORMAT = re.compile(r'^(?P<major>0|[1-9]\d*)\.'
                        r'(?P<minor>0|[1-9]\d*)\.'
                        r'(?P<patch>0|[1-9]\d*)'
                        r'(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][\da-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][\da-zA-Z-]*))*))?'  # noqa
                        r'(?:\+(?P<buildmetadata>[\da-zA-Z-]+(?:\.[\da-zA-Z-]+)*))?$')

    def __init__(self, semver: Union[str, bytes]):
        try:
            parsed = self.FORMAT.match(semver).groupdict()
        except (TypeError, AttributeError) as e:
            # TypeError: semver is not str or bytes
            # AttributeError: invalid semver repr, leads re.match -> NoneType
            raise ValueError(
                'Bad semantic version representation \'%s\'' %
                semver) from e

        self.major = int(parsed['major'])
        self.minor = int(parsed['minor'])
        self.patch = int(parsed['patch'])
        self.pre_release = parsed['prerelease']

    def _cmp(self, sv):
        if self.major != sv.major:
            return self.major - sv.major
        if self.minor != sv.minor:
            return self.minor - sv.minor
        if self.patch != sv.patch:
            return self.patch - sv.patch

        if self.pre_release is None and sv.pre_release is None:
            return 0
        if self.pre_release is None or sv.pre_release is None:
            return 1 if self.pre_release is None else -1

        self_release = self.pre_release.split('.')
        other_release = sv.pre_release.split('.')

        for i in range(min(len(self_release), len(other_release))):
            res = SemVer._cmp_release(self_release[i], other_release[i])
            if res != 0:
                return res

        return len(self_release) - len(other_release)

    @staticmethod
    def _cmp_release(i, o):
        i_int, o_int = True, True
        try:
            i = int(i)
        except ValueError:
            i_int = False
        try:
            o = int(o)
        except ValueError:
            o_int = False

        if i_int and o_int:
            return i - o
        if i_int or o_int:
            return 1 if o_int else -1

        if i == o:
            return 0
        return 1 if i > o else -1

    @peer_cmp_only
    def __eq__(self, other):
        return self._cmp(other) == 0

    @peer_cmp_only
    def __gt__(self, other):
        return self._cmp(other) > 0

    @peer_cmp_only
    def __ge__(self, other):
        return self._cmp(other) >= 0

    @peer_cmp_only
    def __lt__(self, other):
        return self._cmp(other) < 0

    @peer_cmp_only
    def __le__(self, other):
        return self._cmp(other) <= 0



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from featureprobe.hit_result import HitResult
from featureprobe.internal.json_decoder import json_decoder
from featureprobe.model.split import Split

if TYPE_CHECKING:
    from featureprobe.user import User


class Serve:
    def __init__(self,
                 select: int,
                 split: "Split"):
        self._select = select
        self._split = split

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Serve":
        select = json.get('select')
        split = Split.from_json(json.get('split'))
        return cls(select, split)

    @property
    def select(self) -> int:
        return self._select

    @select.setter
    def select(self, value: int):
        self._select = value

    @property
    def split(self) -> "Split":
        return self._split

    @split.setter
    def split(self, value: "Split"):
        self._split = value

    def eval_index(self, user: "User", toggle_key: str) -> "HitResult":
        if self._select is not None:
            return HitResult(hit=True, index=self._select)

        return self._split.find_index(user, toggle_key)



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hashlib import sha1
from typing import List, Optional, TYPE_CHECKING

from featureprobe.hit_result import HitResult
from featureprobe.internal.json_decoder import json_decoder

if TYPE_CHECKING:
    from featureprobe.user import User


class Split:
    _BUCKET_SIZE = 10000

    def __init__(self,
                 distribution: List[List[List[int]]],
                 bucket_by: str,
                 salt: str):
        self._distribution = distribution or []
        self._bucket_by = bucket_by
        self._salt = salt

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Split":
        distribution = json.get('distribution')
        bucket_by = json.get('bucketBy')
        salt = json.get('salt')
        return cls(distribution, bucket_by, salt)

    @property
    def distribution(self) -> List[List[List[int]]]:
        return self._distribution

    @distribution.setter
    def distribution(self, value: List[List[List[int]]]):
        self._distribution = value or []

    @property
    def bucket_by(self) -> str:
        return self._bucket_by

    @bucket_by.setter
    def bucket_by(self, value: str):
        self._bucket_by = value

    @property
    def salt(self) -> str:
        return self._salt

    @salt.setter
    def salt(self, value: str):
        self._salt = value

    def find_index(self, user: "User", toggle_key: str) -> "HitResult":
        hash_key = user.key
        if self._bucket_by:
            if user.has_attr(self._bucket_by):
                hash_key = user.attrs.get(self._bucket_by)
            else:
                return HitResult(
                    hit=False,
                    reason='Warning: User with key \'%s\' does not have attribute name \'%s\'' %
                    (user.key,
                     self._bucket_by))
        group_index = self._get_group(
            self._hash(
                hash_key,
                self._salt or toggle_key,
                self._BUCKET_SIZE))
        return HitResult(hit=True, index=group_index,
                         reason='selected %d percentage group' % group_index)

    def _get_group(self, hash_value: int) -> int:
        for index, groups in enumerate(self._distribution):
            for rng in groups:
                if rng[0] <= hash_value < rng[1]:
                    return index
        return -1  # TODO: return None, HitResult.hit = False? inconsistent with Java SDK

    @staticmethod
    def _hash(hash_key: str, hash_salt: str, bucket_size: int) -> int:
        value = (hash_key + hash_salt).encode('utf-8')
        sha = sha1(value)
        _bytes = sha.digest()[-4:]
        return int.from_bytes(_bytes, byteorder='big') % bucket_size



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import List, Dict, Union, Optional, TYPE_CHECKING

from featureprobe.internal.empty_str import empty_str
from featureprobe.internal.json_decoder import json_decoder
from featureprobe.internal.semver import SemVer
from featureprobe.internal.stringifiable import stringifiable
from featureprobe.model.predicate import ConditionType, Predicate

if TYPE_CHECKING:
    from featureprobe.model.segment import Segment
    from featureprobe.user import User


@stringifiable
class Condition:
    _logger = logging.getLogger('FeatureProbe-Evaluator')

    def __init__(self,
                 subject: str,
                 type_: Union["ConditionType", str],
                 predicate: Union["Predicate", str],
                 objects: List[str]):
        self._subject = subject
        try:
            self._type = ConditionType(type_)
        except ValueError:
            self._type = None
        if isinstance(predicate, Predicate):
            self._predicate = predicate
        elif self._type is not None:
            try:
                self._predicate = self._type.predicates(predicate)
            except ValueError:
                self._predicate = None
        else:
            self._predicate = None
        self._objects = objects or []

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Condition":
        subject = json.get('subject')
        type_ = json.get('type')
        predicate = json.get('predicate')
        objects = json.get('objects')
        return cls(subject, type_, predicate, objects)

    def match_objects(self, user: "User",
                      segments: Optional[Dict[str, "Segment"]]) -> bool:
        if self._type is None or self._predicate is None:
            return False

        matcher_proc = {
            ConditionType.STRING: self._match_string_condition,
            ConditionType.SEGMENT: self._match_segment_condition,
            ConditionType.DATETIME: self._match_datetime_condition,
            ConditionType.NUMBER: self._match_number_condition,
            ConditionType.SEMVER: self._match_semver_condition,
        }

        match = matcher_proc.get(self._type, Condition._match_dummy_condition)
        return match(user=user, segments=segments)

    def _match_string_condition(self, user: "User", **_) -> bool:
        subject_val = user[self._subject]
        if empty_str(subject_val):
            return False
        return self._predicate.matcher(subject_val, self._objects)

    def _match_segment_condition(
            self, user: "User", segments: Dict[str, "Segment"], **_) -> bool:
        return self._predicate.matcher(user, segments or {}, self._objects)

    def _match_datetime_condition(self, user: "User", **_):
        cv = user[self._subject] or time.time()
        try:
            cv = int(cv)
        except ValueError:
            # sourcery skip: replace-interpolation-with-fstring
            self._logger.error(
                'User attribute type mismatch. attribute value: \'%s\', target type int' %
                cv)
            return False

        try:
            return self._predicate.matcher(cv, self._objects)
        except ValueError as e:
            self._logger.error(
                'Met a string that cannot be parsed to int in Condition.objects',
                exc_info=e)
            return False

    def _match_number_condition(self, user: "User", **_):
        cv = user[self._subject]
        if not cv:
            return False
        try:
            cv = float(cv)
        except ValueError:
            # sourcery skip: replace-interpolation-with-fstring
            self._logger.error(
                'User attribute type mismatch. attribute value: \'%s\', target type float' %
                cv)
            return False

        try:
            return self._predicate.matcher(cv, self._objects)
        except ValueError as e:
            self._logger.error(
                'Met a string that cannot be parsed to float in Condition.objects',
                exc_info=e)
            return False

    def _match_semver_condition(self, user: "User", **_):
        cv = user[self._subject]
        try:
            cv = SemVer(cv)
        except ValueError as e:
            # sourcery skip: replace-interpolation-with-fstring
            self._logger.error(
                'Invalid user attribute. attribute value: \'%s\', target type semver' %
                cv, exc_info=e)
            return False

        try:
            return self._predicate.matcher(cv, self._objects)
        except ValueError as e:
            self._logger.error(
                'Met a string that cannot be parsed to semver in Condition.objects',
                exc_info=e)
            return False

    @staticmethod
    def _match_dummy_condition(**_):
        return False

    @property
    def type(self) -> ConditionType:
        return self._type

    @type.setter
    def type(self, value: Union[ConditionType, str]):
        self._type = ConditionType(value)

    @property
    def subject(self) -> str:
        return self._subject

    @subject.setter
    def subject(self, value: str):
        self._subject = value

    @property
    def predicate(self) -> Predicate:
        return self._predicate

    @predicate.setter
    def predicate(self, value: Union[Predicate, str]):
        if isinstance(value, Predicate):
            self._predicate = value
        elif isinstance(value, str) and self._type is not None:
            self._predicate = self._type.predicates(value)

    @property
    def objects(self) -> List[str]:
        return self._objects

    @objects.setter
    def objects(self, value: List[str]):
        self._objects = value



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from featureprobe.model.condition import Condition
from featureprobe.model.distribution import Distribution

from featureprobe.model.predicate import (
    ConditionType,
    StringPredicate,
    SegmentPredicate,
    DatetimePredicate,
    NumberPredicate,
    SemverPredicate,
)

from featureprobe.model.repository import Repository
from featureprobe.model.rule import Rule

from featureprobe.model.segment import (
    Segment,
    SegmentRule,
)

from featureprobe.model.serve import Serve
from featureprobe.model.split import Split
from featureprobe.model.toggle import Toggle

__all__ = [
    'Condition',
    'ConditionType',
    'StringPredicate',
    'SegmentPredicate',
    'DatetimePredicate',
    'NumberPredicate',
    'SemverPredicate',
    'Distribution',
    'Repository',
    'Rule',
    'Segment',
    'SegmentRule',
    'Serve',
    'Split',
    'Toggle',
]



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def defaultable(cls):
    _defaults = {}

    def check_or_create(*args, **kwargs):
        if args or kwargs:
            return cls(*args, **kwargs)
        if cls not in _defaults:
            default = cls()
            _defaults[cls] = default
            return default
        return _defaults[cls]

    return check_or_create



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, TYPE_CHECKING, Dict

from featureprobe.hit_result import HitResult
from featureprobe.internal.json_decoder import json_decoder
from featureprobe.model.condition import Condition
from featureprobe.model.predicate import ConditionType

if TYPE_CHECKING:
    from featureprobe.user import User


class SegmentRule:
    def __init__(self, conditions: List["Condition"] = None):
        self._conditions = conditions or []

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "SegmentRule":
        conditions = [Condition.from_json(c)
                      for c in json.get('conditions', [])]
        return cls(conditions)

    @property
    def conditions(self) -> List["Condition"]:
        return self._conditions

    @conditions.setter
    def conditions(self, value: List["Condition"]):
        self._conditions = value or []

    def hit(self,
            user: "User",
            segments: Dict[str, "Segment"]
            ) -> HitResult:
        for condition in self._conditions:
            if condition.type != ConditionType.SEGMENT and not user.has_attr(
                    condition.subject):
                return HitResult(
                    hit=False, reason="Warning: User with key '%s' does not have attribute name '%s'" %
                    (user.key, condition.subject))
            if not condition.match_objects(user, segments):
                return HitResult(hit=False)

        return HitResult(True)


class Segment:
    def __init__(self,
                 uid: str,
                 version: int,
                 rules: List["SegmentRule"] = None):
        self._uid = uid
        self._version = version
        self._rules = rules or []

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Segment":
        uid = json.get('uniqueId')
        version = json.get('version', 1)
        rules = [SegmentRule.from_json(r) for r in json.get('rules', [])]
        return cls(uid, version, rules)

    @property
    def uid(self):
        return self._uid

    @uid.setter
    def uid(self, value: str):
        self._uid = value

    @property
    def version(self) -> int:
        return self._version

    @version.setter
    def version(self, value: int):
        self._version = value

    @property
    def rules(self) -> List["SegmentRule"]:
        return self._rules

    @rules.setter
    def rules(self, value: List["SegmentRule"]):
        self._rules = value or []

    def contains(self, user: "User", segments: Dict[str, "Segment"]):
        for rule in self._rules:
            hit_result = rule.hit(user, segments)
            if hit_result.hit:
                return True

        return False



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


def json_decoder(func):
    def wrapper(cls, json_):
        if json_ is None:
            return None
        if isinstance(json_, dict):
            return func(cls, json_)

        return func(cls, json.loads(json_))

    return wrapper



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Dict, TYPE_CHECKING

from featureprobe.evaluation_result import EvaluationResult
from featureprobe.internal.json_decoder import json_decoder
from featureprobe.model.rule import Rule
from featureprobe.model.serve import Serve

if TYPE_CHECKING:
    from featureprobe.hit_result import HitResult
    from featureprobe.model.segment import Segment
    from featureprobe.user import User


class Toggle:
    def __init__(self,
                 key: str,
                 enabled: bool,
                 version: int,
                 disabled_serve: "Serve",
                 default_serve: "Serve",
                 rules: List["Rule"],
                 variations: list,
                 for_client: bool):
        self._key = key
        self._enabled = enabled
        self._version = version
        self._disabled_serve = disabled_serve
        self._default_serve = default_serve
        self._rules = rules
        self._variations = variations
        self._for_client = for_client

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Toggle":
        key = json.get('key')
        enabled = json.get('enabled', False)
        version = json.get('version', 1)
        disabled_serve = Serve.from_json(json.get('disabledServe'))
        default_serve = Serve.from_json(json.get('defaultServe'))
        rules = [Rule.from_json(r) for r in json.get('rules', [])]
        variations = json.get('variations', [])
        for_client = json.get('forClient', False)
        return cls(
            key,
            enabled,
            version,
            disabled_serve,
            default_serve,
            rules,
            variations,
            for_client)

    @property
    def key(self) -> str:
        return self._key

    @key.setter
    def key(self, value: str):
        self._key = value

    @property
    def enabled(self) -> bool:
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value

    @property
    def version(self) -> int:
        return self._version

    @version.setter
    def version(self, value: int):
        self._version = value

    @property
    def disabled_serve(self) -> "Serve":
        return self._disabled_serve

    @disabled_serve.setter
    def disabled_serve(self, value: "Serve"):
        self._disabled_serve = value

    @property
    def default_serve(self) -> "Serve":
        return self._default_serve

    @default_serve.setter
    def default_serve(self, value: "Serve"):
        self._default_serve = value

    @property
    def rules(self) -> List["Rule"]:
        return self._rules

    @rules.setter
    def rules(self, value: List["Rule"]):
        self._rules = value or []

    @property
    def variations(self) -> List[str]:
        return self._variations

    @variations.setter
    def variations(self, value: List[str]):
        self._variations = value or []

    @property
    def for_client(self) -> bool:
        return self._for_client

    @for_client.setter
    def for_client(self, value: bool):
        self._for_client = value

    def eval(self,
             user: "User",
             segments: Dict[str,
                            "Segment"],
             default_value: object) -> "EvaluationResult":
        if not self._enabled:
            return self._create_disabled_result(user, self._key, default_value)

        warning = None

        for index, rule in enumerate(self._rules or []):
            hit_result = rule.hit(user, segments, self._key)
            if hit_result.hit:
                return self._hit_value(hit_result, default_value, index)
            warning = hit_result.reason

        return self._create_default_result(
            user, self._key, default_value, warning)

    def _create_disabled_result(
            self,
            user: "User",
            toggle_key: str,
            default_value: object) -> "EvaluationResult":
        disabled_result = self._hit_value(
            self._disabled_serve.eval_index(
                user, toggle_key), default_value)
        disabled_result.reason = 'Toggle disabled'
        return disabled_result

    def _create_default_result(self, user: "User", toggle_key: str,
                               default_value: object,
                               warning: str) -> "EvaluationResult":
        default_result = self._hit_value(
            self._default_serve.eval_index(
                user, toggle_key), default_value)
        # sourcery skip: replace-interpolation-with-fstring
        default_result.reason = 'Default rule hit. %s' % warning
        return default_result

    def _hit_value(self, hit_result: "HitResult", default_value: object,
                   rule_index: Optional[int] = None) -> "EvaluationResult":
        res = EvaluationResult(
            default_value,
            rule_index,
            hit_result.index,
            self._version,
            hit_result.reason or '')
        if hit_result.index is not None:
            variation = self._variations[hit_result.index]
            if isinstance(variation, int) and isinstance(default_value, float):
                res.value = float(variation)
            else:
                res.value = variation
            if rule_index is not None:
                res.reason = 'Rule %d hit' % rule_index

        return res



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

import featureprobe


def _requirements():
    with open('requirements.txt', 'r', encoding='utf-8') as f:
        return [req.strip() for req in f.readlines()]


pypi_classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
]

if 'a' in featureprobe.__version__:
    pypi_classifiers.append('Development Status :: 3 - Alpha')
elif 'b' in featureprobe.__version__ or 'rc' in featureprobe.__version__:
    pypi_classifiers.append('Development Status :: 4 - Beta')
else:
    pypi_classifiers.append('Development Status :: 5 - Production/Stable')

setup(
    name='featureprobe-server-sdk-python',
    version=featureprobe.__version__,

    author='FeatureProbe',
    license='Apache 2.0',
    license_file='LICENSE',

    description='FeatureProbe Server Side SDK for Python',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',

    keywords=['feature management', 'server sdk'],
    classifiers=pypi_classifiers,

    package_dir={'': 'featureprobe'},
    packages=find_packages(where='featureprobe'),

    python_requires='>=3.5, <4',
    install_requires=_requirements(),

    project_urls={
        'Project Homepage': 'https://github.com/FeatureProbe',
        'Source Code': 'https://github.com/FeatureProbe/server-sdk-python',
        'Bug Reports': 'https://github.com/FeatureProbe/server-sdk-python/issues',
    }
)



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict, TYPE_CHECKING

from featureprobe.hit_result import HitResult
from featureprobe.internal.json_decoder import json_decoder
from featureprobe.model.condition import Condition, ConditionType
from featureprobe.model.serve import Serve

if TYPE_CHECKING:
    from featureprobe.model.segment import Segment
    from featureprobe.user import User


class Rule:
    def __init__(self,
                 serve: "Serve" = None,
                 conditions: List["Condition"] = None):
        self._serve = serve
        self._conditions = conditions or []

    @classmethod
    @json_decoder
    def from_json(cls, json: dict) -> "Rule":
        serve = Serve.from_json(json.get('serve'))
        conditions = [Condition.from_json(c)
                      for c in json.get('conditions', [])]
        return cls(serve, conditions)

    @property
    def serve(self) -> "Serve":
        return self._serve

    @serve.setter
    def serve(self, value: "Serve"):
        self._serve = value

    @property
    def conditions(self) -> List["Condition"]:
        return self._conditions

    @conditions.setter
    def conditions(self, value: List["Condition"]):
        self._conditions = value or []

    def hit(self,
            user: "User",
            segments: Dict[str,
                           "Segment"],
            toggle_key: str) -> HitResult:
        for condition in self._conditions:
            if condition.type not in (
                    ConditionType.SEGMENT,
                    ConditionType.DATETIME) and not user.has_attr(
                    condition.subject):
                return HitResult(
                    hit=False, reason="Warning: User with key '%s' does not have attribute name '%s'" %
                    (user.key, condition.subject))
            if not condition.match_objects(user, segments):
                return HitResult(hit=False)

        return self._serve.eval_index(user, toggle_key)



# Copyright 2022 FeatureProbe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from enum import Enum

from featureprobe.internal.semver import SemVer


class Predicate(str, Enum):
    def __new__(cls, value, matcher=None):
        if isinstance(value, cls):
            return value
        if matcher is None:
            return cls._value2member_map_[value]

        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.matcher = matcher
        return obj

    def __str__(self):
        return self.value


class StringPredicate(Predicate):
    IS_ONE_OF = 'is one of', lambda target, objects: \
        target in objects

    ENDS_WITH = 'ends with', lambda target, objects: \
        any(target.endswith(o) for o in objects)

    STARTS_WITH = 'starts with', lambda target, objects: \
        any(target.startswith(o) for o in objects)

    CONTAINS = 'contains', lambda target, objects: \
        any(o in target for o in objects)

    MATCHES_REGEX = 'matches regex', lambda target, objects: \
        any(re.search(pattern=o, string=target) for o in objects)

    IS_NOT_ANY_OF = 'is not any of', lambda target, objects: \
        target not in objects

    DOES_NOT_END_WITH = 'does not end with', lambda target, objects: \
        all(not target.endswith(o) for o in objects)

    DOES_NOT_START_WITH = 'does not start with', lambda target, objects: \
        all(not target.startswith(o) for o in objects)

    DOES_NOT_CONTAIN = 'does not contain', lambda target, objects: \
        all(o not in target for o in objects)

    DOES_NOT_MATCH_REGEX = 'does not match regex', lambda target, objects: \
        all(not re.search(pattern=o, string=target) for o in objects)


class SegmentPredicate(Predicate):
    IS_IN = 'is in', lambda user, segments, objects: \
        any(segments.get(s).contains(user, segments) for s in objects)

    IS_NOT_IN = 'is not in', lambda user, segments, objects: \
        all(not segments.get(s).contains(user, segments) for s in objects)


class DatetimePredicate(Predicate):
    AFTER = 'after', lambda target, objects: \
        any(target >= int(o) for o in objects)

    BEFORE = 'before', lambda target, objects: \
        any(target < int(o) for o in objects)


class NumberPredicate(Predicate):
    EQUAL = '=', lambda custom_value, objects: \
        any(custom_value == float(o) for o in objects)

    NOT_EQUAL = '!=', lambda custom_value, objects: \
        all(custom_value != float(o) for o in objects)

    GREATER_THAN = '>', lambda custom_value, objects: \
        any(custom_value > float(o) for o in objects)

    GREATER_OR_EQUAL = '>=', lambda custom_value, objects: \
        any(custom_value >= float(o) for o in objects)

    LESS_THAN = '<', lambda custom_value, objects: \
        any(custom_value < float(o) for o in objects)

    LESS_OR_EQUAL = '<=', lambda custom_value, objects: \
        any(custom_value <= float(o) for o in objects)


class SemverPredicate(Predicate):
    EQUAL = '=', lambda custom_value, objects: \
        any(custom_value == SemVer(o) for o in objects)

    NOT_EQUAL = '!=', lambda custom_value, objects: \
        all(custom_value != SemVer(o) for o in objects)

    GREATER_THAN = '>', lambda custom_value, objects: \
        any(custom_value > SemVer(o) for o in objects)

    GREATER_OR_EQUAL = '>=', lambda custom_value, objects: \
        any(custom_value >= SemVer(o) for o in objects)

    LESS_THAN = '<', lambda custom_value, objects: \
        any(custom_value < SemVer(o) for o in objects)

    LESS_OR_EQUAL = '<=', lambda custom_value, objects: \
        any(custom_value <= SemVer(o) for o in objects)


class ConditionType(str, Enum):
    def __new__(cls, value, predicates):
        if isinstance(value, cls):
            return value
        if predicates is None:
            return cls._value2member_map_[value]

        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.predicates = predicates
        return obj

    def __str__(self):
        return self.value

    STRING = 'string', StringPredicate
    SEGMENT = 'segment', SegmentPredicate
    DATETIME = 'datetime', DatetimePredicate
    NUMBER = 'number', NumberPredicate
    SEMVER = 'semver', SemverPredicate
